DAT objectName         byte "Hmc5883lCompass140523c", 0     
CON
{{
  ******* Public Notes *******

  HMC5883L Compass

  Written by: Duane Degn
  First posted to forum: May 12, 2014

  For more information about this object see the following
  thread in the Parallax Propeller forum.

  http://forums.parallax.com/showthread.php/155653

  This object is loosely based on the demo object from Parallax.
  Most of the I2C code is from the the original demo object.
  
  In order to provide a stable reading, multiple sensor readings
  are averaged together. The number of readings used in the
  average is determined by the values of the constants
  "BITS_TO_AVERAGE_STATIC" and "BITS_TO_AVERAGE_MOVING".
  These values will be used to determine the size of the
  averaging buffers. Since it's faster to shift bits than
  divide, the buffers must be a power of two. By setting
  the bits in the two previously mentioned constants, the
  buffers are insured to be a power of two value.

  A different number of samples are averaged if the compass
  is determined to be rotating. These fewer samples while
  moving keeps the compass readings from appearing sluggish.

  The averages will be incorrect at start up. The averaging
  buffer needs to fill before the average reading will be
  meaningful.

}}
{
  ******* Private Notes *******

  131118d Start using floating point math.
  18e Working okay.
  18f Try to come up with a more meaningful use of data.
  18f The tilt doesn't seem correct.
  18h Appears to work correctly.
  18i Displays additional debug.
  18j Try using circular buffer.
  18k Sign extend while shifting.
  18k Appears to work in many ways. It doesn't detect
  with the robot is moving.
  18k Moving issue appears to be fixed.
  19a Add a delay so static buffer has time to fill
  with static values prior to being used.
  19a Seems to be working as expected.
  19b Start making changes to separate to demo and
  child objects.
  20b Raw readings seem to be working again.
  20c Seems to be working okay.
  20d Try increasing read rate above the debug's
  ability to keep up.
  20d Seems to work okay with the fast loop.
  140512a Begin work to make presentable enough to
  post to forum.
  13a Add calibration feature.
  16a Add lock so I2C can be shared.
  21c Add a bunch of features to configure the
  sensor.
  21c Not tested yet.
  22a Getting close to how I want it to work.
  22b Remove averaging buffers and other extras.
  23b Add lock parameter to "Init" method. A lock
  could be useful even if the compass isn't being
  monitored by its own cog.
  23c Change name from "Compass140523c" to
  "Hmc5883lCompass140523c"
  
  
}
CON

  WRITE_DATA     = $3C ' Requests Write operation 
  READ_DATA      = $3D ' Requests Read operation

  CONFIG_REG_A = $00
  CONFIG_REG_B = $01
  MODE_REG = $02 ' Mode setting register
  OUTPUT_X_MSB_REG = $03 ' X MSB data output register  
  STATUS_REG = $09
  ID_REG_A = $0A
  ID_REG_B = $0B
  ID_REG_C = $0C

  FIRST_GAIN_BIT = 5
  GAIN_BITS = 3
  FIRST_AVERAGING_BIT = 5
  AVERAGING_BITS = 2
  FIRST_OUTPUT_RATE_BIT = 2
  OUTPUT_RATE_BITS = 3
  FIRST_CONFIG_BIT = 0
  CONFIG_BITS = 2
  FIRST_MODE_BIT = 0
  MODE_BITS = 2
  
  MIN_RAW = -2048
  MAX_RAW = 2047
  
  PSEUDOREAL_MULTIPLIER = 1000  '' user changeable but don't use less than 10 or more than 100,000
  PSEUDO_CIRLE = 360 * PSEUDOREAL_MULTIPLIER 

  F_PSEUDOREAL_MULTIPLIER = float(PSEUDOREAL_MULTIPLIER)

  #0, ONE_SAMPLE_AVERAGED, TWO_SAMPLES_AVERAGED, FOUR_SAMPLES_AVERAGED, EIGHT_SAMPLES_AVERAGED

  MAX_AVERAGING_INDEX = EIGHT_SAMPLES_AVERAGED
  DEFAULT_AVERAGING_INDEX = ONE_SAMPLE_AVERAGED

  #0, NO_CONFIG, POSITIVE_BIAS_CONFIG, NEGATIVE_BIAS_CONFIG, RESERVED_CONFIG

  MAX_CONFIG_INDEX = POSITIVE_BIAS_CONFIG
  DEFAULT_CONFIG_INDEX = NO_CONFIG
  
  #0, GAUSS_1370_GAIN, GAUSS_1090_GAIN, GAUSS_8200_GAIN, GAUSS_660_GAIN, GAUSS_440_GAIN, {
  } GAUSS_390_GAIN, GAUSS_330_GAIN, GAUSS_230_GAIN
  
  MAX_GAIN_INDEX = GAUSS_230_GAIN
  DEFAULT_GAIN_INDEX = GAUSS_1090_GAIN
  
  #0, RATE_0_75_HZ, RATE_1_5_HZ, RATE_3_HZ, RATE_7_5_HZ, RATE_15_HZ, RATE_30_HZ, RATE_75_HZ, RATE_RESERVED_HZ

  MAX_RATE_INDEX = RATE_75_HZ
  DEFAULT_RATE_INDEX = RATE_7_5_HZ
  
  #0, CONTINUOUS_MODE, SINGLE_MODE, IDLE_MODE_0, IDLE_MODE_1

  MAX_MODE_INDEX = IDLE_MODE_1
  DEFAULT_MODE_INDEX = SINGLE_MODE
  
CON

  '' It's faster to shift bits than it is to divide
  '' To find average from buffer total the total
  '' value is shifted by these amounts to find
  '' average.
  '' 1-bit = buffer size of 2
  '' 2-bit = buffer size of 4
  '' 3-bit = buffer size of 8
  '' 4-bit = buffer size of 16
  '' 5-bit = buffer size of 32
  '' 6-bit = buffer size of 64
  '' 7-bit = buffer size of 128
  '' 8-bit = buffer size of 256
  '' Larger buffers will give a more consistent average
  '' but at the cost of taking longer to update to the
  '' new position.
  '' For this reason, I use a different buffer for
  '' moving values.
  '' If the robot makes a heading change of 180 degrees
  '' or more within the time it takes to fill the
  '' moving buffer, the average may be severely
  '' corrupted by the rollover compensations in the
  '' code.
  
  {BITS_TO_AVERAGE_STATIC = 4    '' user changeable  
  BITS_TO_AVERAGE_MOVING = 2    '' user changeable
  BITS_TO_AVERAGE_SPEED = 4     '' user changeable

  SAMPLES_TO_AVERAGE_STATIC = 1 << BITS_TO_AVERAGE_STATIC
  SAMPLES_TO_AVERAGE_MOVING = 1 << BITS_TO_AVERAGE_MOVING
  SAMPLES_TO_AVERAGE_SPEED = 1 << BITS_TO_AVERAGE_SPEED

  MAX_INDEX_STATIC = SAMPLES_TO_AVERAGE_STATIC - 1 
  MAX_INDEX_MOVING = SAMPLES_TO_AVERAGE_MOVING - 1
  MAX_INDEX_SPEED = SAMPLES_TO_AVERAGE_SPEED - 1

  MOVING_TRIGGER = 5            '' user changeable
                                '' in degrees
                                ' difference between two concecutive readings to trigger
                                ' using small (moving) buffer to compute average
                                
  ROLLOVER_ZONE = 90
  ROLLOVER_ZONE_LOW = -ROLLOVER_ZONE
  ROLLOVER_ZONE_HIGH = ROLLOVER_ZONE
  
  BUFFER_SIZE_STATIC = SAMPLES_TO_AVERAGE_STATIC
  BUFFER_SIZE_MOVING = SAMPLES_TO_AVERAGE_MOVING
  BUFFER_SIZE_SPEED = SAMPLES_TO_AVERAGE_SPEED

  INDIVIDUAL_LONGS = 1
  DEVICE_SIZE_ARRAYS = 15
  STATIC_BUFFERS = 2
  MOVING_BUFFERS = 2
  SPEED_BUFFERS = 1

  LONGS_TO_COPY = INDIVIDUAL_LONGS + (DEVICE_SIZE_ARRAYS) + {
  } (STATIC_BUFFERS * BUFFER_SIZE_STATIC) + (MOVING_BUFFERS * BUFFER_SIZE_MOVING) + { 
  } (SPEED_BUFFERS * BUFFER_SIZE_SPEED)
  }
  #0, INDEX_X, INDEX_Y, INDEX_Z
  MAX_SPACE_INDEX = INDEX_Z
  THREE_SPACE_SIZE = MAX_SPACE_INDEX + 1

  #0, HEADING_OFFSET, RAW_MAGNITUDE_OFFSET, MAGNITUDE_OFFSET, TILT_OFFSET

  {  '' Keep variables below together and in order.
  long pseudoHeading, pseudoRawMagnitude, pseudoMagnitude
  long pseudoTilt, pseudoInitialHeading, pseudoInitialTilt
  '' Keep variables above together and in order.
  long[dataPtr][HEADING_OFFSET]
  long[dataPtr][RAW_MAGNITUDE_OFFSET]
  long[dataPtr][MAGNITUDE_OFFSET]
  long[dataPtr][TILT_OFFSET]
  }
  
  CHILD_RAW_OFFSET = 0
  CHILD_COOKED_OFFSET = 3
  CHILD_COUNT_OFFSET = 6
  
  INPUT_HIGH = 0
  OUTPUT_LOW = 1

  GAIN_USED_FOR_EXPECTED_RAW = DEFAULT_GAIN_INDEX
  
  '' The "MAX_EXPECTED_RAW" values were determined with the sensor set at
  '' the same gain index of "GAIN_USED_FOR_EXPECTED_RAW" (defalt value of 1).
  '' These values should be adjusted to match your sensor.
  MAX_EXPECTED_RAW_X = 604   '' user changeable
  MAX_EXPECTED_RAW_Y = 473   '' user changeable
  MAX_EXPECTED_RAW_Z = 610   '' user changeable

  NORMALIZED_COOKED = 1000   '' After data is "cooked" the maximum value of any axis should
                             '' be this amount

  TIMEOUT_INTERVAL = 2_000

VAR ' Child Variables
 
  long stack[128] ' 32 was too small
  long raw[THREE_SPACE_SIZE]
  long cooked[THREE_SPACE_SIZE] 
  long readingCount
  long adjustedMaxExpected[THREE_SPACE_SIZE]
  
  long compassInterval, compassTime
  long dataPtr, i2cLock
 
  long cog, f32Cog, timeoutInterval
    
  byte switchToStaticCount
  byte dataPin, clockPin
  byte idBuffer[4], stopFlag
  byte averagingIndex, rateIndex, configIndex  
  byte gainIndex, modeIndex
  byte readCompassFlag
  
OBJ

  FMath : "F32"
    
PUB Init(clockPin_, dataPin_, dataPtr_, i2cLock_)
  
  clockPin := clockPin_
  dataPin := dataPin_
  dataPtr := dataPtr_

  i2cLock := i2cLock_
    
  f32Cog := FMath.Start
 
  gainIndex := DEFAULT_GAIN_INDEX 
  rateIndex := DEFAULT_RATE_INDEX
  averagingIndex := DEFAULT_AVERAGING_INDEX
  configIndex := DEFAULT_CONFIG_INDEX
  modeIndex := DEFAULT_MODE_INDEX
  SetMaxExpected(gainIndex)
  
  SetMode(CONTINUOUS_MODE)
  
  timeoutInterval := (clkfreq / 1000) + TIMEOUT_INTERVAL

  result := @raw
  
PUB Start(clockPin_, dataPin_, dataPtr_, i2cLock_)
'' This method will start a new cog to continually update the compass
'' readings.
'' Use "Restart" to start a cog to monitor
'' The methods "Start" and "Restart" can cause problems
'' when using other methods of this object.
'' If a cog is used to continually update the
'' heading and tilt variables, then is should be
'' stopped prior to using methods which require
'' I2C interactions with the sensor.

  if cog or f32cog
    Terminate

  Init(clockPin_, dataPin_, dataPtr_, i2cLock_)
  
  cog := cognew(mainLoop, @stack) + 1
  result := @raw

PUB RequestStop
'' This method will set a flag indicating
'' to the cog actively monitoring the
'' compass to shutdown.
'' Having the cog stop itself will insure
'' the I2C lines are properly released.
'' The cog running the F32 object will
'' remain active.

  ifnot cog
    return -1
  result := cog
  stopFlag := 1
  repeat while stopFlag
  waitcnt(clkfreq / 100 + cnt)
  
PRI Harakiri
'' By allowing the cog to stop itself, we can
'' make sure the last communication with
'' the sensor is completed.

  cog := 0
  stopFlag := 0
  cogstop(cogID)

PUB Terminate
'' Use this method to stop a cog which may be
'' monitoring the compass and to stop the cog
'' used for floating point calculations.

  result := RequestStop
  FMath.Stop
  f32Cog := 0
  if i2cLock <> -1
    lockret(i2cLock)
    
PUB GetCog

  result := cog
  
PUB Restart
'' Use this method to restart a cog to
'' actively monitor the compass sensor.
'' The "Init" method needs to have been
'' called prior to calling this method.

  ifnot f32Cog
    return 0
    
  ifnot cog
    cog := cognew(mainLoop, @stack) + 1

PRI MainLoop
 
  repeat
    if stopFlag
      Harakiri
    CheckCompass
    
PUB CheckCompass

  if cnt - compassTime > compassInterval  ' ** use a better way to time reads
    compassTime += compassInterval
    result := ReadCompass

PUB WaitForCompass | localTime

  localTime := cnt
  repeat
    result := ReadStatus
  while result & 2 and cnt - localTime < timeoutInterval
  result := ReadCompass

PUB ReadCompass | localIndex

  ifnot readCompassFlag ' only set pointer if needed
    SetPointer(OUTPUT_X_MSB_REG)
  GetRaw                      ' Gather raw data from compass       
  ProcessCompass

  readingCount++
  result := 1
  repeat localIndex from 0 to MAX_SPACE_INDEX
    if raw[localIndex] < MIN_RAW or raw[localIndex] > MAX_RAW
      result := -1
  
  readCompassFlag := 1

PUB ReadRegisters(firstRegister, registersToRead, pointerToParentRam)
  
  if (firstRegister + registersToRead) + 1 => STATUS_REG and firstRegister < STATUS_REG
    SetPointer(firstRegister)
    StartI2c
    Send(READ_DATA)
    result := (STATUS_REG - firstRegister) - 1
    repeat result
      byte[pointerToParentRam++] := Receive(true)
      
    registersToRead -= result + 1    
    byte[pointerToParentRam++] := Receive(false)
    StopI2c
    firstRegister := STATUS_REG

  registersToRead--   
  SetPointer(firstRegister)
  StartI2c
  Send(READ_DATA)
  repeat registersToRead
    byte[pointerToParentRam++] := Receive(true)
  byte[pointerToParentRam] := Receive(false)
  StopI2c
  
PUB GetCompassId(pointerToParentBuffer)
'' The "Init" or "Start" method should be called prior to
'' calling this method.
'' Returns a pointer to the three character string read
'' from the compass.
'' This method will temporarily stop the compass reading
'' cog. If you wish to keep the cog stopped, call the
'' "Stop" method prior to calling this method.

  ReadRegisters(ID_REG_A, 3, pointerToParentBuffer)
  byte[pointerToParentBuffer][3] := 0

  result := pointerToParentBuffer

PUB GetGainIndex

  result := gainIndex 
  
PUB GetGain

  result := gainResolution[gainIndex] 
  
PUB SetGain(newGainIndex) 

  gainIndex := 0 #> newGainIndex <# MAX_GAIN_INDEX

  SetBits(CONFIG_REG_B, FIRST_GAIN_BIT, GAIN_BITS, gainIndex)

  SetMaxExpected(gainIndex)
  
  result := GetGain

PUB SetMaxExpected(newGainIndex)

  if newGainIndex == GAIN_USED_FOR_EXPECTED_RAW ' default index = 1
    adjustedMaxExpected[INDEX_X] := MAX_EXPECTED_RAW_X
    adjustedMaxExpected[INDEX_Y] := MAX_EXPECTED_RAW_Y
    adjustedMaxExpected[INDEX_Z] := MAX_EXPECTED_RAW_Z
  else                              
    adjustedMaxExpected[INDEX_X] := (MAX_EXPECTED_RAW_X * gainResolution[GAIN_USED_FOR_EXPECTED_RAW]) / {
                                  } gainResolution[newGainIndex]
    adjustedMaxExpected[INDEX_Y] := (MAX_EXPECTED_RAW_Y * gainResolution[GAIN_USED_FOR_EXPECTED_RAW]) / {
                                  } gainResolution[newGainIndex]
    adjustedMaxExpected[INDEX_Z] := (MAX_EXPECTED_RAW_Z * gainResolution[GAIN_USED_FOR_EXPECTED_RAW]) / {
                                  } gainResolution[newGainIndex]
     
PUB GetRateIndex

  result := rateIndex 
  
PUB GetRate

  result := rateHz[rateIndex] 
  
PUB SetRate(newRateIndex)

  rateIndex := 0 #> newRateIndex <# MAX_RATE_INDEX

  SetBits(CONFIG_REG_A, FIRST_OUTPUT_RATE_BIT, OUTPUT_RATE_BITS, rateIndex)
  
  result := GetRate
   
PUB GetAveragingIndex

  result := averagingIndex 
  
PUB GetAveragingNumber

  result := internalAveraging[averagingIndex] 
  
PUB SetAveraging(newAveragingIndex)

  averagingIndex := 0 #> newAveragingIndex <# MAX_AVERAGING_INDEX

  SetBits(CONFIG_REG_A, FIRST_AVERAGING_BIT, AVERAGING_BITS, averagingIndex)

  result := GetAveragingNumber
   
PUB GetConfigIndex

  result := configIndex 
  
PUB SetConfig(newConfigIndex)

  configIndex := 0 #> newConfigIndex <# MAX_CONFIG_INDEX

  SetBits(CONFIG_REG_A, FIRST_CONFIG_BIT, CONFIG_BITS, configIndex)

  result := configIndex
   
PUB GetModeIndex
  
  result := modeIndex
  
PUB SetMode(newModeIndex)

  modeIndex := 0 #> newModeIndex <# MAX_MODE_INDEX

  SetBits(MODE_REG, FIRST_MODE_BIT, MODE_BITS, modeIndex)

  result := modeIndex

PUB GetObjectName

  result := @objectName

PUB GetExpectedId

  result := @expectedIdText

{PUB SetState(newState) ' **** Not ready



PUB SetSensitivity(newSensitivity) ' **** Not ready



PUB SetHome(homeDelay) ' **** Not ready

  waitcnt((clkfreq / 1000) * homeDelay + cnt)
  

PUB SetActiveParameter(dimension, address, minLimit, maxLimit, endType) ' **** Not ready

  activeParameterPtr[dimension] := address
  minParameter[dimension] := minLimit
  maxParameter[dimension] := maxLimit
  parameterEndType[dimension] := endType
  }
PUB ReadStatus

  SetPointer(STATUS_REG)
  StartI2c
  Send(READ_DATA)
  result := receive(false)
  StopI2c

PUB SetBits(register, firstBit, numberOfBits, value) | mask
  ' Sets compass to continuous output mode

  mask := 0
  
  repeat numberOfBits
    mask <<= 1
    mask += 1
  mask <<= firstBit

  SetPointer(register)
  StartI2c
  Send(READ_DATA)
  result := receive(false)
  StopI2c

  result &= !mask
  value <<= firstBit
  result |= value
    
  StartI2c
  send(WRITE_DATA)
  send(register)
  send(result)
  StopI2c

PUB ComfirmSettings(fixBadRegisterFlag) | tempBuffer, tempValue, indicator

  ReadRegisters(CONFIG_REG_A, 3, @tempBuffer)
  indicator := 1
  
  tempValue := IsolateBits(FIRST_AVERAGING_BIT, AVERAGING_BITS, byte[@tempBuffer])
  if tempValue <> averagingIndex
    result += indicator
    if fixBadRegisterFlag
      SetAveraging(averagingIndex)
  indicator *= 10
  tempValue := IsolateBits(FIRST_OUTPUT_RATE_BIT, OUTPUT_RATE_BITS, byte[@tempBuffer])
  if tempValue <> rateIndex
    result += indicator
    if fixBadRegisterFlag
      SetRate(rateIndex)
  indicator *= 10
  tempValue := IsolateBits(FIRST_CONFIG_BIT, CONFIG_BITS, byte[@tempBuffer])
  if tempValue <> configIndex
    result += indicator
    if fixBadRegisterFlag
      SetConfig(configIndex)
  indicator *= 10
  tempValue := IsolateBits(FIRST_GAIN_BIT, GAIN_BITS, byte[@tempBuffer + 1])
  if tempValue <> gainIndex
    result += indicator
    if fixBadRegisterFlag
      SetGain(gainIndex)
  indicator *= 10
  tempValue := IsolateBits(FIRST_MODE_BIT, MODE_BITS, byte[@tempBuffer + 2])
  if tempValue <> modeIndex
    result += indicator
    if fixBadRegisterFlag
      SetMode(modeIndex)
  indicator *= 10

  GetCompassId(@tempBuffer)
  if byte[@tempBuffer] <> expectedIdText[0]
    result += indicator
  indicator *= 10
  if byte[@tempBuffer + 1] <> expectedIdText[1]
    result += indicator 
  indicator *= 10
  if byte[@tempBuffer + 2] <> expectedIdText[2]
    result++    
    
PUB IsolateBits(firstBit, numberOfBits, value) | mask
'' Find the value of the bits of interest in "value"
'' (normally a register value).

  mask := 0
  
  repeat numberOfBits
    mask <<= 1
    mask += 1
  value >>= firstBit

  result := value & mask

PUB SetPointer(register)
'' Start pointer at user specified register
'' The pointer will automatically reset to
'' "OUTPUT_X_MSB_REG" after the last compass
'' reading regiser (#8) is read.

  StartI2c
  Send(WRITE_DATA)
  Send(register)
  StopI2c

PRI GetRaw | tempRaw[THREE_SPACE_SIZE]
  ' Get raw data from continous output
  
  StartI2c
  Send(READ_DATA)
  tempRaw[INDEX_X] := ((receive(true) << 8) | receive(true))
  tempRaw[INDEX_Z] := ((receive(true) << 8) | receive(true))
  tempRaw[INDEX_Y] := ((receive(true) << 8) | receive(false)) 
 
  StopI2c
  repeat result from 0 to MAX_SPACE_INDEX
    raw[result] := ~~tempRaw[result]       
    ' sign extend (change 16-bit number to a 32-bit number
    ' The top 16 bits will be incorrect for negative numbers
    ' the "~~" operator fixes this problem
 
PRI CookCompass

  cooked[INDEX_X] := (raw[INDEX_X] * NORMALIZED_COOKED) / adjustedMaxExpected[INDEX_X]
  cooked[INDEX_Y] := (raw[INDEX_Y] * NORMALIZED_COOKED) / adjustedMaxExpected[INDEX_Y]
  cooked[INDEX_Z] := (raw[INDEX_Z] * NORMALIZED_COOKED) / adjustedMaxExpected[INDEX_Z]

PRI ProcessCompass | fVector[THREE_SPACE_SIZE], fSum, fSumXY, fZ

  CookCompass
  
  repeat result from 0 to MAX_SPACE_INDEX
    fVector[result] := FMath.FFloat(cooked[result])

  fZ := fVector[INDEX_Z]
  
  long[dataPtr][HEADING_OFFSET] := FMath.FRound(FMath.FMul(FMath.Degrees( {
  } FMath.ATan2(fVector[INDEX_Y], fVector[INDEX_X])), F_PSEUDOREAL_MULTIPLIER))
  
  repeat result from 0 to MAX_SPACE_INDEX
    fVector[result] := FMath.FMul(fVector[result], fVector[result])
    if result
      fSum := FMath.FAdd(fSum, fVector[result])
    else  ' set use first index (zero) as initial value of fSum
      fSum := fVector[result]
    if result == INDEX_Y
      fSumXY := fSum  ' save sum of x and y to use later
           
  long[dataPtr][MAGNITUDE_OFFSET] := FMath.FRound(FMath.FMul(FMath.FSqr(fSum), F_PSEUDOREAL_MULTIPLIER))
  long[dataPtr][TILT_OFFSET] := FMath.FRound(FMath.FMul(FMath.Degrees( {
  } FMath.ATan2(fZ, FMath.FSqr(fSumXY))), F_PSEUDOREAL_MULTIPLIER))

  '' compute raw magnitude
  repeat result from 0 to MAX_SPACE_INDEX
    fVector[result] := FMath.FFloat(raw[result])

  repeat result from 0 to MAX_SPACE_INDEX
    fVector[result] := FMath.FMul(fVector[result], fVector[result])
    if result
      fSum := FMath.FAdd(fSum, fVector[result])
    else  ' set use first index (zero) as initial value of fSum
      fSum := fVector[result]
           
  long[dataPtr][RAW_MAGNITUDE_OFFSET] := FMath.FRound(FMath.FMul(FMath.FSqr(fSum), F_PSEUDOREAL_MULTIPLIER))

PUB Send(value)
'' Part of I2C communication.
'' Copied from Gordon McComb's
'' original code.

  value := ((!value) >< 8)

  repeat 8
    dira[dataPin]  := value
    dira[clockPin] := INPUT_HIGH ' clock pulled high
    dira[clockPin] := OUTPUT_LOW
    value >>= 1

  dira[dataPin]  := INPUT_HIGH
  dira[clockPin] := INPUT_HIGH
  result         := !(ina[dataPin])
  dira[clockPin] := OUTPUT_LOW
  dira[dataPin]  := OUTPUT_LOW

PUB StartI2c
'' Set I2C pins as outputs in the low state.

  'repeat while lockset(i2cLock)
  if i2cLock <> -1 
    repeat until not lockset(i2cLock)
  outa[dataPin]  := INPUT_HIGH
  outa[clockPin] := INPUT_HIGH
  dira[dataPin]  := OUTPUT_LOW ' data transitions low while clock is high to indicate start
  dira[clockPin] := OUTPUT_LOW

PUB Receive(aknowledge)
'' Part of I2C communication.
'' Copied from Gordon McComb's
'' original code.

  dira[dataPin] := INPUT_HIGH ' may be driven low by slave device

  repeat 8
    result <<= 1
    dira[clockPin] := INPUT_HIGH
    result |= ina[dataPin]
    dira[clockPin] := OUTPUT_LOW

  dira[dataPin]  := aknowledge
  dira[clockPin] := INPUT_HIGH
  dira[clockPin] := OUTPUT_LOW
  dira[dataPin]  := OUTPUT_LOW

PUB StopI2c

  dira[clockPin] := INPUT_HIGH
  dira[dataPin] := INPUT_HIGH  ' stop condition data pulled high while clock is high
  if i2cLock <> -1 
    lockclr(i2cLock)
  readCompassFlag := 0
  
DAT

rateHz                  long 750, 1_500, 3_000, 7_500, 15_000, 30_000, 75_000
'gainGauss               long 1370, 1090, 8200, 660, 440, 390, 330, 230
gainResolution          long 730, 920, 1_220, 1_520, 2_270, 2_560, 3_030, 4_350
internalAveraging       long 1, 2, 4, 8

expectedIdText          byte "H43", 0       