DAT programName         byte "Hmc5883lDemo140523e", 0     
CON
{{
  ******* Public Notes *******

  HMC5883L Compass

  See notes in child object.

  This demo allows one to test all the features of
  the HMC5883L digital compass.

  A terminal program is required to interact with
  this object.

  Follow prompts on the terminal window to access
  the various features of the compass.

  For additional information see following thread in
  Parallax Propeller forum.

  http://forums.parallax.com/showthread.php/155653
  
}}
{
  ******* Private Notes *******

  131118d Start using floating point math.
  18e Working okay.
  18f Try to come up with a more meaningful use of data.
  18f The tilt doesn't seem correct.
  18h Appears to work correctly.
  18i Displays additional debug.
  18j Try using circular buffer.
  18k Sign extend while shifting.
  18k Appears to work in many ways. It doesn't detect
  with the robot is moving.
  18k Moving issue appears to be fixed.
  19a Add a delay so static buffer has time to fill
  with static values prior to being used.
  19a Seems to be working as expected.
  19b Start making changes to separate to demo and
  child objects.
  20b Raw readings seem to be working again.
  20c Seems to be working okay.
  20d Try increasing read rate above the debug's
  ability to keep up.
  20d Seems to work okay with the fast loop.
  140512a Begin work to make presentable enough to
  post to forum.
  Change name to "CompassDemo140512a". Start
  separating out child object.
  12b Change name to "CompassBasicDemo140512a"
  12a Remove averaging and rollover code.
  12b Remove interval check from child object.
  12c Some small changes to child object. Added
  magnitude to display data.
  12c Raw returned all -1. Not working.
  12d Revert to version 12b of child object.
  13b Add calibration algorithm to child object.
  16a Keep begining tilt in memory.
  22a Use child object with many more features.
  23c z-axis tilt displays -361 or so when
  initialized to heading -178. Problem with
  roll over feature.
  23e Roll over issues appear to be fixed.
  Remove "ROLLOVER_ANGLE" from this
  version.
  23e Change name from "CompassBasicDemo140523e"
  to "DemoHmc5883l140523e"
}
CON ' Parent Constants

  _clkmode      = xtal1 + pll16x
  _clkfreq      = 80_000_000
                                                        
  DATA_PIN  = 29                 '' user changeable                                 
  CLOCK_PIN = 28                 '' user changeable
 
  READINGS_TO_SAMPLE = 32
  ALLOWED_MOVEMENT = 2 * Compass#PSEUDOREAL_MULTIPLIER
  MAX_INITIAL_TILT_ATTEMPTS = 200

  COMPASS_OUTPUT_ROWS = 12
  TOP_OF_MENU = COMPASS_OUTPUT_ROWS
  BOTTOM_OF_MENU =  26
  
VAR ' Parent Variables

  '' Keep variables below together and in order.
  long pseudoHeading, pseudoRawMagnitude, pseudoMagnitude
  long pseudoTilt, pseudoInitialHeading, pseudoInitialTilt
  '' Keep variables above together and in order.

  long childDataPtr, i2cLock
  
OBJ

  Pst : "Parallax Serial Terminal"
  Compass : "Hmc5883lCompass140523c"
  
PUB Demo

  Pst.start(115200)

  i2cLock := locknew ' This lock isn't really needed for this program.
                     ' The lock is useful when using other I2C devices
                     ' on the same I2C bus.
 
  Pst.Clear
  repeat
    waitcnt(clkfreq / 4 + cnt) 
    result := Pst.RxCount
    Pst.str(string(11, 13, " Press any key to continue."))
  until result

  Pst.RxFlush
  
  Pst.Clear
  Pst.str(string(11, 13, " Initializing Compass ")) 
  childDataPtr := Compass.Init(CLOCK_PIN, DATA_PIN, @pseudoHeading, i2cLock)

  Pst.str(string(11, 13, " The compass has been started. Checking compass settings."))

  ComfirmSettings
      
  Pst.str(string(11, 13, " Setting comfirmed. Setting initial gain, heading and tilt."))
    
  InitializeGainHeadingAndTilt(@pseudoInitialHeading, READINGS_TO_SAMPLE, ALLOWED_MOVEMENT, ALLOWED_MOVEMENT, {
} MAX_INITIAL_TILT_ATTEMPTS, Compass#DEFAULT_GAIN_INDEX)

  Pst.str(string(11, 13, " Initialization complete."))

  DisplayMenu
  Loop
   
PUB Loop
'' The "Loop" will read the compass from this cog (#0). 
 
  repeat
    result := Pst.RxCount
    if result
      DisplayMenu
      GetInput      
    Compass.WaitForCompass
    CompassDebug
    TiltDebug
    
PUB GetInput

  result := Pst.CharIn
  Pst.Position(0, BOTTOM_OF_MENU)
  Pst.ClearEnd
  Pst.ClearBelow
  
  case result
    "c", "C":
      Pst.str(string(11, 13, " Clearing Screen"))
      waitcnt(clkfreq / 2 + cnt)
      Pst.Clear
    "g", "G":
      AdjustGain
    "a", "A":
      AdjustAveraging
    "r", "R":
      AdjustRate
    "m", "M":
      AdjustMode
    "t", "T":
      Pst.str(string(11, 13, 11, 13, " Testing Registers", 11, 13))
      ComfirmSettings
    "i", "I":
      ReadId
    "s", "S":
      FindSteady(@pseudoInitialHeading, READINGS_TO_SAMPLE, ALLOWED_MOVEMENT, ALLOWED_MOVEMENT, {
      } MAX_INITIAL_TILT_ATTEMPTS)
    "p", "P":
      AdjustBias(Compass#POSITIVE_BIAS_CONFIG)
    "n", "N":
      AdjustBias(Compass#NEGATIVE_BIAS_CONFIG)  
    "d", "D":
      'Pst.Clear
      DisplayAll
      waitcnt(clkfreq * 3 + cnt)   
    
PUB DisplayMenu

  Pst.Position(0, TOP_OF_MENU)
  Pst.ClearEnd
  Pst.ClearBelow

  Pst.str(string(11, 13, " Press one of the following keys:"))
  Pst.str(string(11, 13, "C) Clear screen"))
  Pst.str(string(11, 13, "G) adjust Gain"))
  Pst.str(string(11, 13, "A) change Averaging used"))
  Pst.str(string(11, 13, "R) change sampling Rate"))
  Pst.str(string(11, 13, "M) change Mode"))
  Pst.str(string(11, 13, "T) Test registers"))
  Pst.str(string(11, 13, "I) read Id registers of compass"))
  Pst.str(string(11, 13, "S) Set initial heading and tilt"))
  Pst.str(string(11, 13, "P) set Positive bias configuration"))
  Pst.str(string(11, 13, "N) set Negative bias configuration"))
  Pst.str(string(11, 13, "D) Display all settings"))
  
PUB CompassDebug 

  Pst.Position(0, 0)

  Pst.str(string(11, 13, " programName = "))
  Pst.str(@programName)
  Pst.str(string(11, 13, " objectName = "))
  Pst.str(Compass.GetObjectName)
  
  Pst.str(string(11, 13, " raw X = "))        
  Pst.dec(long[childDataPtr][0])
  Pst.str(string(", Y = "))
  Pst.dec(long[childDataPtr][1])
  Pst.str(string(", Z = "))
  Pst.dec(long[childDataPtr][2])

  Pst.str(string(11, 13, " cooked X = "))        
  Pst.dec(long[childDataPtr][3])
  Pst.str(string(", Y = "))
  Pst.dec(long[childDataPtr][4])
  Pst.str(string(", Z = "))
  Pst.dec(long[childDataPtr][5])
  Pst.ClearEnd
  Pst.Char(13)
  Pst.str(string(11, 13, " course = "))

  DecPoint(pseudoHeading, Compass#PSEUDOREAL_MULTIPLIER)
 
  Pst.ClearEnd
  Pst.Char(13)

  Pst.str(string(11, 13, " field angle from horizontal = "))
  DecPoint(pseudoTilt, Compass#PSEUDOREAL_MULTIPLIER)    

  Pst.str(string(11, 13, " raw field strength = "))
  DecPoint(pseudoRawMagnitude, Compass#PSEUDOREAL_MULTIPLIER)     
  Pst.str(string(11, 13, " normalized field strength = "))
  DecPoint(pseudoMagnitude, Compass#PSEUDOREAL_MULTIPLIER)
  
PUB TiltDebug

  Pst.str(string(11, 13, " z-axis tilt = "))
  DecPoint(pseudoTilt - pseudoInitialTilt, Compass#PSEUDOREAL_MULTIPLIER)
  Pst.ClearEnd

  Pst.Char(13)
  Pst.ClearEnd

PUB InitializeGainHeadingAndTilt(initalPtrs, readingsToAverage, allowedHeadingMovement, allowedTiltMovement, {
} maxAttempts, minGainIndex) | gainIndex

  Pst.str(string(11, 13, " Initializing Gain"))
  
  DisplayGain
  
  if gainIndex <> minGainIndex
    SetAndDisplayGain(minGainIndex)
    gainIndex := minGainIndex
  
  
  repeat
    result := FindSteady(initalPtrs, readingsToAverage, allowedHeadingMovement, {
                       } allowedTiltMovement, maxAttempts)
    if result == -1
      SetAndDisplayGain(++gainIndex)
  while result == -1 and gainIndex < Compass#MAX_GAIN_INDEX

  if result == -1 or result == Compass#PSEUDO_CIRLE
    Pst.str(string(11, 13, " ****** ERROR ******")) 
    Pst.str(string(11, 13, " Problem with compass."))
    Pst.str(string(11, 13, " Using current heading and tilt as initial state."))
    Compass.CheckCompass 
    long[initalPtrs][0] := pseudoHeading
    long[initalPtrs][1] := pseudoTilt
    
PUB FindSteady(initalPtrs, readingsToAverage, allowedHeadingMovement, allowedTiltMovement, {
} maxAttempts) | readingsSampled, tiltTotal, failedAttempts, minHeading, {
} maxHeading, minTilt, maxTilt, {
} localIndex, headingTotal, rolloverFlag[2], previousAveragingIndex, {
} previousRateIndex, previousModeIndex, localHeading
'' Returns 1 if initialization was successful.
'' Returns 0 if there was too much movements during initialization.
'' Returns -1 if sensor readers were outside normal range.
'' Values will only be stored for initial heading and tilt is this
'' initializaion ran successfully.
 
  failedAttempts := 0
 
  Pst.str(string(11, 13, " Finding Initial Heading and Tilt")) 
  Pst.str(string(11, 13, " Please keep compass stationary."))
  Pst.str(string(11, 13, " Setting compass to use averaging.")) 
  Pst.str(string(11, 13, " 8 samples will be used for each reading."))

  previousAveragingIndex := Compass.GetAveragingIndex
  if previousAveragingIndex <> Compass#EIGHT_SAMPLES_AVERAGED
    Compass.SetAveraging(Compass#EIGHT_SAMPLES_AVERAGED)
  else
    Pst.str(string(11, 13, " The compass was already set to use 8 samples."))  

  Pst.str(string(11, 13, " Setting compass to read at 75 Hz.")) 

  previousRateIndex := Compass.GetRateIndex
  if previousRateIndex <> Compass#RATE_75_HZ
    Compass.SetRate(Compass#RATE_75_HZ)
  else
    Pst.str(string(11, 13, " The compass was already set to read at 75 Hz."))  

  previousModeIndex := Compass.GetModeIndex
  if previousModeIndex <> Compass#CONTINUOUS_MODE
    Compass.SetMode(Compass#CONTINUOUS_MODE)
    Pst.str(string(11, 13, " Setting compass to continuous mode."))
    
  Pst.str(string(11, 13, " Setting initial heading and tilt in 2 seconds."))
  waitcnt(clkfreq + cnt)
  Pst.str(string(11, 13, " Setting initial heading and tilt in 1 second."))
  waitcnt(clkfreq + cnt)
  Pst.str(string(11, 13, " Please keep compass stationary."))

  'previousCount := long[childDataPtr][Compass#CHILD_COUNT_OFFSET]
  'if freezeCount <> previousCount
    'previousCount := freezeCount
  
  
  repeat
    Compass.WaitForCompass 'CheckCompass
    if ||pseudoHeading > (180 * Compass#PSEUDOREAL_MULTIPLIER) - allowedHeadingMovement 
      rolloverFlag[0] := 1
      localHeading := Rollover(pseudoHeading)
    tiltTotal := 0
    headingTotal := 0
    readingsSampled := 0
    maxHeading := minHeading := localHeading
    
    minTilt := maxTilt := pseudoTilt
    Pst.str(string(11, 13, " Start Heading = "))
    DecPoint(minTilt, Compass#PSEUDOREAL_MULTIPLIER) 
    Pst.str(string(11, 13, " Start Tilt = "))
    DecPoint(minTilt, Compass#PSEUDOREAL_MULTIPLIER) 

    repeat
      Compass.WaitForCompass
      if rolloverFlag
        localHeading := Rollover(pseudoHeading)
      else
        localHeading := pseudoHeading
              
      repeat localIndex from 0 to Compass#MAX_SPACE_INDEX
        Pst.dec(long[childDataPtr][localIndex])
        if localIndex <> Compass#MAX_SPACE_INDEX
          Pst.str(string(", "))
        
        if long[childDataPtr][localIndex] < Compass#MIN_RAW or long[childDataPtr][localIndex] > Compass#MAX_RAW
          Pst.str(string(11, 13, " Raw out of range."))
          waitcnt(clkfreq * 2 + cnt)
          RestoreRegisters(previousAveragingIndex, previousRateIndex, previousModeIndex)
          return -1
                  

      Pst.str(string(11, 13, " localHeading = "))
      Pst.dec(localHeading)
      if localHeading < minHeading
        minHeading := localHeading
        Pst.str(string(11, 13, " new minHeading = "))
        DecPoint(minHeading, Compass#PSEUDOREAL_MULTIPLIER)
       
        Pst.ClearEnd  
      elseif localHeading > maxHeading
        maxHeading := localHeading
        Pst.str(string(11, 13, " new maxHeading = "))
        DecPoint(maxHeading, Compass#PSEUDOREAL_MULTIPLIER)
       
        Pst.ClearEnd
      if maxHeading - minHeading > allowedHeadingMovement
        Pst.str(string(11, 13, " Too much movement detected. allowedHeadingMovement = "))
        DecPoint(allowedHeadingMovement, Compass#PSEUDOREAL_MULTIPLIER)
    
        Pst.ClearEnd
        Pst.str(string(11, 13, " movement while initializing = "))
        DecPoint(maxHeading - minHeading, Compass#PSEUDOREAL_MULTIPLIER)
  
        Pst.ClearEnd
        failedAttempts++
        Pst.char(7) ' bell
        quit

      if pseudoTilt < minTilt
        minTilt := pseudoTilt
        Pst.str(string(11, 13, " new minTilt = "))
        DecPoint(minTilt, Compass#PSEUDOREAL_MULTIPLIER)
        Pst.ClearEnd  
      elseif pseudoTilt > maxTilt
        minTilt := pseudoTilt
        Pst.str(string(11, 13, " new maxTilt = "))
        DecPoint(maxTilt, Compass#PSEUDOREAL_MULTIPLIER)
        Pst.ClearEnd
      if maxTilt - minTilt > allowedTiltMovement
        Pst.str(string(11, 13, " Too much movement detected. allowedMovement = "))
        DecPoint(allowedTiltMovement, Compass#PSEUDOREAL_MULTIPLIER)
        Pst.ClearEnd
        Pst.str(string(11, 13, " movement while initializing = "))
        DecPoint(maxTilt - minTilt, Compass#PSEUDOREAL_MULTIPLIER)
        Pst.ClearEnd
        failedAttempts++
        Pst.char(7) ' bell
        quit
      readingsSampled++
   
      tiltTotal += pseudoTilt     
      headingTotal += localHeading
        
    until readingsSampled == readingsToAverage        
  until readingsSampled == readingsToAverage or failedAttempts == maxAttempts

  Pst.str(string(11, 13, " headingTotal = "))
  Pst.dec(headingTotal)
  Pst.str(string(11, 13, " tiltTotal = "))
  Pst.dec(tiltTotal)

  result := headingTotal / readingsSampled
  result := LimitAngle(result)  
  long[initalPtrs][0] := result
 
  long[initalPtrs][1] := tiltTotal / readingsSampled

  Pst.str(string(11, 13, " initial heading = "))
  DecPoint(long[initalPtrs][0], Compass#PSEUDOREAL_MULTIPLIER)  
  Pst.ClearEnd
  Pst.str(string(11, 13, " initial tilt = "))
  DecPoint(long[initalPtrs][1], Compass#PSEUDOREAL_MULTIPLIER)
  Pst.ClearEnd

  RestoreRegisters(previousAveragingIndex, previousRateIndex, previousModeIndex)
      
  if readingsSampled == readingsToAverage
    result := 1
    Pst.char(7) ' bell
    waitcnt(clkfreq / 6 + cnt)
    Pst.char(7) ' bell 

  else
    result := 0
    Pst.char(7) ' bell
    waitcnt(clkfreq / 6 + cnt)
    Pst.char(7) ' bell
    waitcnt(clkfreq / 6 + cnt)
    Pst.char(7) ' bell
    waitcnt(clkfreq / 6 + cnt)
    Pst.char(7) ' bell 

  PauseForKeyPress
  
PUB PauseForKeyPress

  Pst.str(string(11, 13, " Press any key to continue."))
  repeat
    result := Pst.RxCount
  until result

  Pst.RxFlush
  
PUB RestoreRegisters(previousAveragingIndex, previousRateIndex, previousModeIndex)

  if previousAveragingIndex <> Compass#EIGHT_SAMPLES_AVERAGED   
    Pst.str(string(11, 13, " The compass will now be set back to averaging "))
    Pst.dec(Compass.SetAveraging(previousAveragingIndex))
    Pst.str(string(" samples."))
    
  Pst.str(string(11, 13, " Setting compass to read at 75 Hz.")) 

  if previousRateIndex <> Compass#RATE_75_HZ
    Pst.str(string(11, 13, " The compass will now be set back to to reading at "))
    DecPoint(Compass.SetRate(previousRateIndex), Compass#PSEUDOREAL_MULTIPLIER)
    Pst.str(string(" Hz"))     
  else
    Pst.str(string(11, 13, " The compass will remain set to read at 75 Hz."))    

  if previousModeIndex <> Compass#CONTINUOUS_MODE
    Compass.SetMode(previousModeIndex)
    Pst.str(string(11, 13, " Returning compass to mode # "))
    Pst.dec(Compass.SetMode(previousModeIndex))

PUB SetAndDisplayGain(newIndex)

  result := Compass.SetGain(newIndex)
  Pst.str(string(11, 13, " gain changed to ")) 
  DecPoint(result, Compass#PSEUDOREAL_MULTIPLIER)
  Pst.str(string(" milli-Gauss per count (index = ")) 
  Pst.dec(newIndex)  
  Pst.char(")")
    
PUB ComfirmSettings

  result := Compass.ComfirmSettings(true)

  if result
    Pst.str(string(11, 13, " The compass settings were not those expected."))
    Pst.str(string(11, 13, " ComfirmSettings returned ", 34))
    Pst.dec(result)
    Pst.char(34)
    Pst.str(string(11, 13, " Checking compass settings again."))
    result := Compass.ComfirmSettings(true)
    if result
      Pst.str(string(11, 13, " The compass settings are still not those expected."))
      Pst.str(string(11, 13, " ComfirmSettings returned ", 34))
      Pst.dec(result)
      Pst.char(34)
      Pst.str(string(11, 13, " ****** ERROR ******"))
      Pst.str(string(11, 13, " Check connections to compass and try again."))
    else
      Pst.str(string(11, 13, " The compass registers have been properly set."))   
  else
    Pst.str(string(11, 13, " The compass registers held the expected values."))
    
PUB AdjustGain

  DisplayGain

  Pst.str(string(11, 13, " Please enter new gain index."))
  result := Pst.DecIn
  Compass.SetGain(result)       

  DisplayGain

PUB AdjustAveraging

  DisplayAveraging

  Pst.str(string(11, 13, " Please enter new averaging index."))
  result := Pst.DecIn
  Compass.SetAveraging(result)
        
  DisplayAveraging

  
PUB AdjustRate

  DisplayRate

  Pst.str(string(11, 13, " Please enter new rate index."))
  result := Pst.DecIn
  Compass.SetRate(result)       

  DisplayRate

PUB AdjustMode

  DisplayMode

  Pst.str(string(11, 13, " Please enter new mode index."))
  result := Pst.DecIn
  Compass.SetMode(result)       

  DisplayMode

PUB AdjustBias(bias)

  Pst.str(string(11, 13, " A current to bias the three sensors on the compass will now be turned one."))        
  Pst.str(string(11, 13, " The raw data displayed below may be used to assist in calibrationg the sensor."))        
  Pst.str(string(11, 13, " If possible, record the temperature of the sensor at the time this data is collected."))

  DisplayAll

  case bias
    Compass#POSITIVE_BIAS_CONFIG:
      Pst.str(string(11, 13, 11, 13, " Postitive Bias Active")) 
    Compass#NEGATIVE_BIAS_CONFIG:
      Pst.str(string(11, 13, 11, 13, " Negative Bias Active"))

  waitcnt(clkfreq + cnt)
  Pst.str(string(11, 13, 11, 13, " Press any key to return to previous menu."))        
  waitcnt(clkfreq + cnt)
  Compass.SetConfig(bias)
  
  repeat
    Compass.WaitForCompass
    result := Pst.RxCount
    Pst.str(string(11, 13, " raw X = "))        
    Pst.dec(long[childDataPtr][0])
    Pst.str(string(", Y = "))
    Pst.dec(long[childDataPtr][1])
    Pst.str(string(", Z = "))
    Pst.dec(long[childDataPtr][2])
  
  until result

  Pst.str(string(11, 13, " Turning off bias."))        

  Compass.SetConfig(Compass#NO_CONFIG)
  waitcnt(clkfreq + cnt)
  
  Pst.RxFlush

PUB DisplayGain

  Pst.str(string(11, 13, " current gain = ")) 
  result := Compass.GetGain
  DecPoint(result, Compass#PSEUDOREAL_MULTIPLIER)
  Pst.str(string(" milli-Gauss per count (index = ")) 
  result := Compass.GetGainIndex
  Pst.dec(result)  
  Pst.char(")")

PUB DisplayAveraging
  
  Pst.str(string(11, 13, " current samples averaged = ")) 
  result := Compass.GetAveragingNumber
  Pst.dec(result)
  Pst.str(string(" (index = ")) 
  result := Compass.GetAveragingIndex
  Pst.dec(result)  
  Pst.char(")")

PUB DisplayRate
       
  Pst.str(string(11, 13, " current smampling rate = ")) 
  result := Compass.GetRate
  DecPoint(result, Compass#PSEUDOREAL_MULTIPLIER)
  Pst.str(string(" readings per second (index = ")) 
  result := Compass.GetRateIndex
  Pst.dec(result)  
  Pst.char(")")

PUB DisplayMode

  Pst.str(string(11, 13, " current mode index = ")) 
  result := Compass.GetModeIndex
  Pst.dec(result)
  Pst.ClearEnd
  Pst.char(13)
  Pst.char(" ")

  case result
    Compass#CONTINUOUS_MODE:
      Pst.str(string("continuous read mode"))
    Compass#SINGLE_MODE:
      Pst.str(string("single read mode"))  
    Compass#IDLE_MODE_0, Compass#IDLE_MODE_1:
      Pst.str(string("idle mode"))  
  
PUB DisplayAll

  DisplayGain
  
  DisplayAveraging
       
  DisplayRate

  DisplayMode

PUB ReadId | tempBuffer

  Pst.str(string(13, " Compass ID = ", 34))
  Compass.GetCompassId(@tempBuffer)
  SafeStr(@tempBuffer, 3, false)
  Pst.Char(34)

PUB Rollover(localAngle)

  if localAngle < 0
    result := localAngle + Compass#PSEUDO_CIRLE
  else
    result := localAngle
    
PUB LimitAngle(localAngle)

  if localAngle < -180 * Compass#PSEUDOREAL_MULTIPLIER
    result := localAngle + Compass#PSEUDO_CIRLE
  elseif localAngle > 180 * Compass#PSEUDOREAL_MULTIPLIER
    result := localAngle - Compass#PSEUDO_CIRLE
  else
    result := localAngle
    
PUB DecPoint(value, denominator)

  if value < 0
    Pst.Char("-")
    -value
      
  if value => denominator
    result := value / denominator
    Pst.Dec(result)
    value //= denominator     
  else    
    Pst.Char("0")
  Pst.Char(".")  
  repeat while denominator > 1
    denominator /= 10
    if value => denominator
      result := value / denominator
      Pst.Dec(result)
      value //= denominator
    else
      Pst.Char("0")

PUB SafeStr(stringPtr, size, stopOnZeroFlag)

  repeat size
    if stopOnZeroFlag and byte[stringPtr] == 0  
      quit ' break out of loop 
    case byte[stringPtr]
      " ".."~": ' Printable ASCII characters
        Pst.Char(byte[stringPtr])
      other: ' Not a printable character so display as hex.
        Pst.Char("<")
        Pst.Char("$")
        Pst.Hex(byte[stringPtr], 2)
        Pst.Char(">")
    stringPtr++

{PUB NewLine

  Pst.ClearEnd
  Pst.Newline
    }